"""Streaming"""

from __future__ import annotations
import codecs

from baretypes import Content


async def bytes_reader(content: Content) -> bytes:
    """Extracts the body content as bytes.

    Args:
        content (Content): The content argument of the request handler.

    Returns:
        bytes: The body as bytes.
    """
    buf = b''
    async for b in content:
        buf += b
    return buf


async def text_reader(content: Content, encoding: str = 'utf-8') -> str:
    """Extracts the body contents as text.

    Args:
        content (Content): The content argument of the request handler.
        encoding (str, optional): The encoding of the text. Defaults to 'utf-8'.

    Returns:
        str: The body contents as a string.
    """
    codec_info: codecs.CodecInfo = codecs.lookup(encoding)
    decoder = codec_info.incrementaldecoder()
    text = ''
    async for b in content:
        text += decoder.decode(b)
    return text


async def bytes_writer(buf: bytes, chunk_size: int = -1) -> Content:
    """Creates an asynchronous iterator from the supplied response body.

    Args:
        buf (bytes): The response body to return.
        chunk_size (int, optional): The size of each chunk to send or -1 to send
            as a single chunk.. Defaults to -1.

    Yields:
        Content: The body bytes
    """
    if chunk_size == -1:
        yield buf
    else:
        start, end = 0, chunk_size
        while start < len(buf):
            yield buf[start:end]
            start, end = end, end + chunk_size


async def text_writer(
        text: str,
        encoding: str = 'utf-8',
        chunk_size: int = -1
) -> Content:
    """Creates an asynchronous iterator from the supplied response body.

    Args:
        text (str): The response body.
        encoding (str, optional): The encoding to apply when transforming the
            text into bytes. Defaults to 'utf-8'.
        chunk_size (int, optional): The size of each chunk to send or -1 to send
            as a single chunk.. Defaults to -1.

    Yields:
        Content: The body bytes
    """
    if chunk_size == -1:
        yield text.encode(encoding=encoding)
    else:
        start, end = 0, chunk_size
        while start < len(text):
            yield text[start:end].encode(encoding=encoding)
            start, end = end, end + chunk_size



"""Unpackers"""

from email.parser import BytesFeedParser
from email.message import Message
from typing import List, Tuple, Union

from baretypes import Content

MessageParams = List[Tuple[str, str]]
MessagePayload = Union[List[Message], str, bytes, None]


async def unpack_multipart_form_data(
        content_type: bytes,
        content: Content
) -> List[Tuple[MessageParams, MessagePayload]]:
    """Unpack multipart form data

    Args:
        content_type (bytes): The 'content-type' header
        content (Content): The content to parse.

    Raises:
        AssertionError: When the problems were found with the data

    Returns:
        List[Tuple[MessageParams, MessagePayload]]: The form and files
    """
    # Create the parser and prime it with the content
    # type.
    parser = BytesFeedParser()
    parser.feed(b'Content-Type: ')  # type: ignore[arg-type]
    parser.feed(content_type)  # type: ignore[arg-type]
    parser.feed(b'\r\n\r\n')  # type: ignore[arg-type]

    # Feed the content to the parser.
    async for buf in content:
        parser.feed(buf)
    msg = parser.close()

    assert msg is not None, "The parser should return a message"
    assert msg.is_multipart(), "The message should be a multipart message"
    assert not msg.defects, "The message should not have defects"

    msg_payload = msg.get_payload()
    assert msg_payload is not None, "The message payload should not be empty"

    data: List[Tuple[MessageParams, MessagePayload]] = []
    for msg_part in msg_payload:
        assert isinstance(
            msg_part, Message), "A message part should also be a message"
        assert not msg_part.is_multipart(), "A message part should not be multipart"
        assert not msg_part.defects, "A message part should not have defects"

        content_disposition = msg_part.get_params(header='content-disposition')
        assert content_disposition, "A message part should have a content-disposition header"

        part_payload: MessagePayload = msg_part.get_payload(decode=True)
        data.append(
            (content_disposition, part_payload)
        )

    return data



"""Header utilities

A collection of functions to extract headers from the ASGI scope.
"""

import collections
from datetime import datetime
from enum import Enum, auto
from typing import List, Optional, Mapping, MutableMapping, Any, Tuple, Callable, NamedTuple

from baretypes import Headers

from .cookies import decode_cookies, decode_set_cookie
from .dates.rfc_7231 import parse_date


class _MergeType(Enum):
    NONE = auto()
    EXTEND = auto()
    APPEND = auto()
    CONCAT = auto()


class _Parser(NamedTuple):
    parse: Callable[[bytes], Any]
    merge_type: _MergeType


_PARSERS: MutableMapping[bytes, _Parser] = dict()


def _pass_through(value: bytes) -> bytes:
    return value


_DEFAULT_PARSER = _Parser(_pass_through, _MergeType.APPEND)


def index(name: bytes, headers: Headers) -> int:
    """Find the index of the header in the list.

    Args:
        name (bytes): The header name.
        headers (Headers): The headers to search.

    Returns:
        int: The index of the header or -1 if not found.
    """
    return next((i for i, (k, v) in enumerate(headers) if k == name), -1)


def find(name: bytes, headers: Headers, default: Optional[bytes] = None) -> Optional[bytes]:
    """Find the value of a header, or return a default value.

    Args:
        name (bytes): The header name.
        headers (Headers): The headers to search.
        default (Optional[bytes], optional): An optional default value. Defaults
            to None.

    Returns:
        Optional[bytes]: The value of the header if found, otherwise the default value.
    """
    return next((v for k, v in headers if k == name), default)


def find_all(name: bytes, headers: Headers) -> List[bytes]:
    """Find all the values for a given header.

    Args:
        name (bytes): The header name.
        headers (Headers): The headers to search.

    Returns:
        List[bytes]: A list of the header values which may be empty if there
            were none found.
    """
    return [v for k, v in headers if k == name]


def upsert(name: bytes, value: bytes, headers: Headers) -> None:
    """If the header exists overwrite the value, otherwise append a new value.

    Args:
        name (bytes): The header name.
        value (bytes): The header value.
        headers (Headers): The headers.
    """
    for i, item in enumerate(headers):
        if item[0] == name:
            headers[i] = (name, value)
            return
    headers.append((name, value))


def to_dict(headers: Headers) -> MutableMapping[bytes, List[bytes]]:
    """Convert a list of headers into a dictionary where the key is the header
    name and the value is a list of the values of the headers for that name

    Args:
        headers (Headers): A list of headers.

    Returns:
        MutableMapping[bytes, List[bytes]]: A dictionary where the key is the
            header name and the value is a list of the values of the headers for
            that name
    """
    items: MutableMapping[bytes, List[bytes]] = collections.defaultdict(list)
    for name, value in headers:
        items[name].append(value)
    return items


def _parse_date(value: bytes) -> datetime:
    return parse_date(value.decode())


def find_date(
        name: bytes,
        headers: Headers,
        *,
        default: Optional[datetime] = None
) -> Optional[datetime]:
    """Find a header containing a date.

    Args:
        name (bytes): The name of the header.
        headers (Headers): The headers.
        default (Optional[datetime], optional): The headers, Defaults to None.

    Returns:
        Optional[datetime]: The date if found, otherwise the default value.
    """
    value = find(name, headers)
    return default if not value else _parse_date(value)


def _parse_comma_separated_list(value: bytes) -> List[bytes]:
    return [item.strip() for item in value.split(b',')]


def _parse_int(value: bytes) -> int:
    return int(value)


def _parse_float(value: bytes) -> float:
    return float(value)


def _parse_quality(value: bytes) -> Optional[float]:
    if value == b'':
        return None
    name, quality = value.split(b'=')
    if name != b'q':
        raise ValueError('expected "q"')
    return float(quality)


def _parse_accept_quality(value: bytes) -> Tuple[bytes, Any]:
    if value == b'':
        return b'q', 1.0
    name, quality = value.split(b'=')
    return name, float(quality) if name == b'q' else quality


def _parse_media_type_and_encoding(
        value: bytes
) -> Tuple[bytes, Optional[bytes]]:
    media_type, sep, rest = value.partition(b';')
    if not sep:
        encoding = None
    else:
        tag, sep, encoding = rest.partition(b'=')
        if tag != b'charset':
            raise Exception('encoding must start with chartset')
    return media_type.strip(), encoding.strip() if encoding else None


# Accept

def _parse_accept(
        value: bytes,
        *,
        add_wildcard: bool = False
) -> Mapping[bytes, Tuple[bytes, Any]]:
    values = {
        first: _parse_accept_quality(rest)
        for first, sep, rest in [x.strip().partition(b';') for x in value.split(b',')]
    }

    if add_wildcard and b'*' not in values:
        values[b'*'] = (b'q', 1.0)

    return values


_PARSERS[b'accept'] = _Parser(_parse_accept, _MergeType.NONE)


def accept(
        headers: Headers,
        *,
        add_wildcard: bool = False,
        default: Optional[Mapping[bytes, Tuple[bytes, Any]]] = None
) -> Optional[Mapping[bytes, Tuple[bytes, Any]]]:
    """Returns the accept header if it exists.

    Where quality is not given it defaults to 1.0.

    ```python
    >>> accept([(b'accept', b'text/html, application/xhtml+xml, application/xml;q=0.9, */*;q=0.8')])
    {b'text/html': 1.0, b'application/xhtml+xml': 1.0, b'application/xml': 0.9, b'*/*': 0.8}
    ```

    Args:
        headers (Headers): The headers
        add_wildcard (bool, optional): If true add the implicit wildcard '*'.
            Defaults to False.
        default (Optional[Mapping[bytes, float]], optional): An optional
            default. Defaults to None.

    Returns:
        Optional[Mapping[bytes, Tuple[bytes, Any]]]: A dictionary where the key
            is media type and the value is quality.
    """
    value = find(b'accept', headers)
    return default if value is None else _parse_accept(value, add_wildcard=add_wildcard)

# Accept-CH


_PARSERS[b'accept-ch'] = _Parser(_parse_comma_separated_list,
                                 _MergeType.CONCAT)


def accept_ch(
        headers: Headers,
        *,
        default: Optional[List[bytes]] = None
) -> Optional[List[bytes]]:
    """The Accept-CH header is set by the server to specify which Client Hints
    headers client should include in subsequent requests.

    Args:
        headers (Headers): The headers
        default (Optional[Mapping[bytes, float]], optional): An optional
            default. Defaults to None.

    Returns:
        Optional[List[bytes]]: The client hints
    """
    value = find(b'accept-ch', headers)
    return default if value is None else _parse_comma_separated_list(value)

# Accept-CH-Lifetime


_PARSERS[b'accept-ch-lifetime'] = _Parser(_parse_int, _MergeType.NONE)


def accept_ch_lifetime(
        headers: Headers,
        *,
        default: Optional[int] = None
) -> Optional[int]:
    """The Accept-CH-Lifetime header is set by the server to specify the
    persistence of Accept-CH header value that specifies for which Client Hints
    headers client should include in subsequent requests.

    Args:
        headers (Headers): The headers
        default (Optional[Mapping[bytes, float]], optional): An optional
            default. Defaults to None.

    Returns:
        Optional[int]: The lifetime in seconds
    """
    value = find(b'accept-ch-lifetime', headers)
    return default if value is None else _parse_int(value)


# Accept-Charset

def _parse_accept_charset(value: bytes, *, add_wildcard: bool = False) -> Mapping[bytes, float]:
    charsets = {
        first: _parse_quality(rest) or 1.0
        for first, sep, rest in [x.partition(b';') for x in value.split(b', ')]
    }

    if add_wildcard and b'*' not in charsets:
        charsets[b'*'] = 1.0

    return charsets


_PARSERS[b'accept-charset'] = _Parser(_parse_accept_charset, _MergeType.NONE)


def accept_charset(
        headers: Headers,
        *,
        add_wildcard: bool = False,
        default: Optional[Mapping[bytes, float]] = None
) -> Optional[Mapping[bytes, float]]:
    """Extracts the accept encoding header if it exists into a mapping of the
    encoding and the quality value which defaults to 1.0 if missing.

    Args:
        headers (Headers): The headers
        add_wildcard (bool, optional): If True ensures the '*' charset is
            included. Defaults to False.
        default (Optional[Mapping[bytes, float]], optional): An optional
            default. Defaults to None.

    Returns:
        Optional[Mapping[bytes, float]]: A mapping of the encodings and qualities.
    """
    value = find(b'accept-charset', headers)
    return default if value is None else _parse_accept_charset(
        value,
        add_wildcard=add_wildcard
    )

# Accept-Encoding


def _parse_accept_encoding(value: bytes, *, add_identity: bool = False) -> Mapping[bytes, float]:
    encodings = {
        first: _parse_quality(rest) or 1.0
        for first, sep, rest in [x.partition(b';') for x in value.split(b', ')]
    }

    if add_identity and b'identity' not in encodings:
        encodings[b'identity'] = 1.0

    return encodings


_PARSERS[b'accept-encoding'] = _Parser(_parse_accept_encoding, _MergeType.NONE)


def accept_encoding(
        headers: Headers,
        *,
        add_identity: bool = False,
        default: Optional[Mapping[bytes, float]] = None
) -> Optional[Mapping[bytes, float]]:
    """Extracts the accept encoding header if it exists into a mapping of the encoding
    and the quality value which defaults to 1.0 if missing.

    Args:
        headers (Headers): The headers to search.
        add_identity (bool, optional): If True ensures the 'identity' encoding
            is included.. Defaults to False.
        default (Optional[Mapping[bytes, float]], optional): An optional
            default. Defaults to None.

    Returns:
        Optional[Mapping[bytes, float]]: A mapping of the encodings and qualities.
    """
    value = find(b'accept-encoding', headers)
    return default if value is None else _parse_accept_encoding(value, add_identity=add_identity)


# Accept-Language

def _parse_accept_language(value: bytes, *, add_wildcard: bool = False) -> Mapping[bytes, float]:
    languages = {
        first: _parse_quality(rest) or 1.0
        for first, sep, rest in [x.partition(b';') for x in value.split(b', ')]
    }

    if add_wildcard and b'*' not in languages:
        languages[b'*'] = 1.0

    return languages


_PARSERS[b'accept-language'] = _Parser(_parse_accept_language, _MergeType.NONE)


def accept_language(
        headers: Headers,
        *,
        add_wildcard: bool = False,
        default: Optional[Mapping[bytes, float]] = None
) -> Optional[Mapping[bytes, float]]:
    """Extracts the accept language header if it exists into a mapping of the
    encoding and the quality value which defaults to 1.0 if missing.

    Args:
        headers (Headers): The headers to search.
        add_wildcard (bool, optional): If True ensures the '*' charset is
            included. Defaults to False.
        default (Optional[Mapping[bytes, float]], optional): [description].
            Defaults to None.

    Returns:
        Optional[Mapping[bytes, float]]: A mapping of the encodings and
            qualities.
    """
    value = find(b'accept-language', headers)
    return default if value is None else _parse_accept_language(value, add_wildcard=add_wildcard)


# Accept-Patch

def _parse_accept_patch(value: bytes) -> List[Tuple[bytes, Optional[bytes]]]:
    return [
        _parse_media_type_and_encoding(item)
        for item in value.split(b',')
    ]


_PARSERS[b'accept-patch'] = _Parser(_parse_accept_encoding, _MergeType.NONE)


def accept_patch(
        headers: Headers,
        *,
        default: Optional[List[Tuple[bytes, Optional[bytes]]]] = None
) -> Optional[List[Tuple[bytes, Optional[bytes]]]]:
    """The Accept-Patch response HTTP header advertises which media-type the
    server is able to understand.

    Args:
        headers (Headers): The headers to search.
        default (Optional[List[Tuple[bytes, Optional[bytes]]]], optional): An
            optional default value. Defaults to None.

    Returns:
        Optional[List[Tuple[bytes, Optional[bytes]]]]: A list of tuples of media
            type and optional charset.
    """
    value = find(b'accept-patch', headers)
    return default if value is None else _parse_accept_patch(value)


# Accept-Ranges

def _parse_accept_ranges(value: bytes) -> bytes:
    return value.strip()


_PARSERS[b'accept-ranges'] = _Parser(_parse_accept_ranges, _MergeType.NONE)


def accept_ranges(
        headers: Headers,
        *,
        default: Optional[bytes] = None
) -> Optional[bytes]:
    """Returns the value of the accept ranges header of None if missing

    Args:
        headers (Headers): The headers
        default (Optional[bytes], optional): An optional default value. Defaults
            to None.

    Returns:
        Optional[bytes]: The header value (bytes or none)
    """
    value = find(b'accept-ranges', headers)
    return default if value is None else _parse_accept_ranges(value)


# Access-Control-Allow-Credentials

def _parse_access_control_allow_credentials(value: bytes) -> bool:
    return value.lower() == b'true'


_PARSERS[b'access-control-allow-credentials'] = _Parser(
    _parse_access_control_allow_credentials, _MergeType.NONE)


def access_control_allow_credentials(
        headers: Headers,
        *,
        default: Optional[bool] = None
) -> Optional[bool]:
    """Extracts the access control allow credentials header as a bool or None if
    missing.

    Args:
        headers (Headers): The headers.
        default (Optional[bool], optional): An optional default value. Defaults
            to None.

    Returns:
        Optional[bool]: A bool or None
    """
    value = find(b'access-control-allow-credentials', headers)
    return default if value is None else _parse_access_control_allow_credentials(value)


# Access-Control-Allow-Headers

_PARSERS[b'access-control-allow-headers'] = _Parser(
    _parse_comma_separated_list, _MergeType.NONE)


def access_control_allow_headers(
        headers: Headers,
        *,
        default: Optional[List[bytes]] = None
) -> Optional[List[bytes]]:
    """The Access-Control-Allow-Headers response header is used in response to
    a preflight request which includes the Access-Control-Request-Headers to
    indicate which HTTP headers can be used during the actual request.

    Args:
        headers (Headers): The headers.
        default (Optional[List[bytes]], optional): An optional default value.
            Defaults to None.

    Returns:
        Optional[List[bytes]]: A list of the allowed headers or '*' for all headers.
    """
    value = find(b'access-control-allow-headers', headers)
    return default if value is None else _parse_comma_separated_list(value)


# Access-Control-Allow-Methods

_PARSERS[b'access-control-allow-methods'] = _Parser(
    _parse_comma_separated_list, _MergeType.NONE)


def access_control_allow_methods(
        headers: Headers,
        *,
        default: Optional[List[bytes]] = None
) -> Optional[List[bytes]]:
    """The Access-Control-Allow-Methods response header specifies the method or
    methods allowed when accessing the resource in response to a preflight
    request.

    Args:
        headers (Headers): The headers.
        default (Optional[List[bytes]], optional): An optional default value.
            Defaults to None.

    Returns:
        Optional[List[bytes]]: A list of the allowed methods, or '*' for all
            methods.
    """
    value = find(b'access-control-allow-methods', headers)
    return default if value is None else _parse_comma_separated_list(value)


# Access-Control-Allow-Origin

_PARSERS[b'access-control-allow-origin'] = _Parser(
    _pass_through, _MergeType.NONE)


def access_control_allow_origin(
        headers: Headers,
        *,
        default: Optional[bytes] = None
) -> Optional[bytes]:
    """The Access-Control-Allow-Origin response header indicates whether the
    response can be shared with requesting code from the given origin.

    Args:
        headers (Headers): The headers.
        default (Optional[List[bytes]], optional): An optional default value.
            Defaults to None.

    Returns:
        Optional[bytes]: The origin or '*' for all origins, or 'null'
    """
    value = find(b'access-control-allow-origin', headers)
    return default if value is None else value


# Access-Control-Expose-Headers

def _parse_access_control_expose_headers(
        value: bytes,
        *,
        add_simple_response_headers: bool = False
) -> List[bytes]:
    headers = _parse_comma_separated_list(value)
    if add_simple_response_headers:
        headers.extend([
            b'cache-control',
            b'content-language',
            b'content-type',
            b'expires',
            b'last-modified',
            b'pragma',
        ])
    return headers


_PARSERS[b'access-control-expose-headers'] = _Parser(
    _parse_access_control_expose_headers, _MergeType.NONE)


def access_control_expose_headers(
        headers: Headers,
        *,
        add_simple_response_headers: bool = False,
        default: Optional[List[bytes]] = None
) -> Optional[List[bytes]]:
    """[summary]

    Args:
        headers (Headers): The headers.
        add_simple_response_headers (bool, optional): If true add the safelisted
            headers. Defaults to False.
        default (Optional[List[bytes]], optional): An optional default value.
            Defaults to None.

    Returns:
        Optional[List[bytes]]: The headers to expose.
    """
    value = find(b'access-control-expose-headers', headers)
    return default if value is None else _parse_access_control_expose_headers(
        value,
        add_simple_response_headers=add_simple_response_headers
    )

# Access-Control-Max-Age


_PARSERS[b'access-control-max-age'] = _Parser(_parse_int, _MergeType.NONE)


def access_control_max_age(
        headers: Headers,
        *,
        default: Optional[int] = None
) -> Optional[int]:
    """The Access-Control-Max-Age response header indicates how long the results
    of a preflight request (that is the information contained in the
    Access-Control-Allow-Methods and Access-Control-Allow-Headers headers) can
    be cached.

    Args:
        headers (Headers): The headers
        default (Optional[int], optional): An optional default value. Defaults
            to None.

    Returns:
        Optional[int]: The number of seconds
    """
    value = find(b'access-control-max-age', headers)
    return default if value is None else _parse_int(value)

# Access-Control-Request-Headers


_PARSERS[b'access-control-request-headers'] = _Parser(
    _parse_comma_separated_list, _MergeType.NONE)


def access_control_request_headers(
        headers: Headers,
        *,
        default: Optional[List[bytes]] = None
) -> Optional[List[bytes]]:
    """The Access-Control-Request-Headers request header is used by browsers
    when issuing a preflight request, to let the server know which HTTP headers
    the client might send when the actual request is made.

    Args:
        headers (Headers): The headers
        default (Optional[List[bytes]], optional): An optional default value.
            Defaults to None.

    Returns:
        Optional[List[bytes]]: The request headers
    """
    value = find(b'access-control-request-headers', headers)
    return default if value is None else _parse_comma_separated_list(value)

# Access-Control-Request-Method


_PARSERS[b'access-control-request-method'] = _Parser(
    _pass_through, _MergeType.NONE)


def access_control_request_method(
        headers: Headers,
        *,
        default: Optional[bytes] = None
) -> Optional[bytes]:
    """The Access-Control-Request-Method request header is used by browsers when
    issuing a preflight request, to let the server know which HTTP method will
    be used when the actual request is made. This header is necessary as the
    preflight request is always an OPTIONS and doesn't use the same method as
    the actual request.

    Args:
        headers (Headers): The headers.
        default (Optional[bytes], optional): An optional default value. Defaults
            to None.

    Returns:
        Optional[bytes]: The method
    """
    value = find(b'access-control-request-method', headers)
    return default if value is None else _pass_through(value)

# Age


_PARSERS[b'age'] = _Parser(_parse_int, _MergeType.NONE)


def age(
        headers: Headers,
        *,
        default: Optional[int] = None
) -> Optional[int]:
    """The Age header contains the time in seconds the object has been in a
    proxy cache.

    Args:
        headers (Headers): The headers.
        default (Optional[int], optional): An optional default value. Defaults to None.

    Returns:
        Optional[int]: The time in seconds.
    """
    value = find(b'age', headers)
    return default if value is None else _parse_int(value)

# Allow


_PARSERS[b'allow'] = _Parser(_parse_comma_separated_list, _MergeType.NONE)


def allow(
        headers: Headers,
        *,
        default: Optional[List[bytes]] = None
) -> Optional[List[bytes]]:
    """The Allow header lists the set of methods supported by a resource.

    Args:
        headers (Headers): The headers.
        default (Optional[List[bytes]], optional): An optional default value.
            Defaults to None.

    Returns:
        Optional[List[bytes]]: A list of methods
    """
    value = find(b'allow', headers)
    return default if value is None else _parse_comma_separated_list(value)


def _parse_authorization(value: bytes) -> Tuple[bytes, bytes]:
    auth_type, _, credentials = value.partition(b' ')
    return auth_type.strip(), credentials


_PARSERS[b'authorization'] = _Parser(_parse_authorization, _MergeType.NONE)


def authorization(
        headers: Headers,
        *,
        default: Optional[Tuple[bytes, bytes]] = None
) -> Optional[Tuple[bytes, bytes]]:
    """The HTTP Authorization request header contains the credentials to
    authenticate a user agent with a server, usually after the server has
    responded with a 401 Unauthorized status and the WWW-Authenticate header.

    Args:
        headers (Headers): The headers.
        default (Optional[Tuple[bytes, bytes]], optional): An optional default
            value. Defaults to None.

    Returns:
        Optional[Tuple[bytes, bytes]]: The type and credentials.
    """
    value = find(b'authorization', headers)
    return default if value is None else _parse_authorization(value)

# Cache-Control


def _parse_cache_control(value: bytes) -> Mapping[bytes, Optional[int]]:
    return {
        name.strip(): int(rest) if sep == b'=' else None
        for name, sep, rest in [item.partition(b'=') for item in value.split(b',')]
    }


_PARSERS[b'cache-control'] = _Parser(_parse_cache_control, _MergeType.NONE)


def cache_control(
        headers: Headers,
        *,
        default: Optional[Mapping[bytes, Optional[int]]] = None
) -> Optional[Mapping[bytes, Optional[int]]]:
    """The Cache-Control general-header field is used to specify directives for
    caching mechanisms in both requests andresponses. Caching directives are
    unidirectional, meaning that a given directive in a request is not implying
    that the same directive is to be given in the response.

    Args:
        headers (Headers): The headers.
        default (Optional[Mapping[bytes, Optional[int]]], optional): An optional
            default value. Defaults to None.

    Returns:
        Optional[Mapping[bytes, Optional[int]]]: A dictionary of the directives
            and values.
    """
    value = find(b'cache-control', headers)
    return default if value is None else _parse_cache_control(value)


# Clear-Site-Data

_PARSERS[b'clear-site-data'] = _Parser(
    _parse_comma_separated_list, _MergeType.NONE)


def clear_site_data(
        headers: Headers,
        *,
        default: Optional[List[bytes]] = None
) -> Optional[List[bytes]]:
    """The Clear-Site-Data header clears browsing data (cookies, storage, cache)
    associated with the requesting website. It allows web developers to have
    more control over the data stored locally by a browser for their origins.

    Args:
        headers (Headers): The headers.
        default (Optional[List[bytes]], optional): An optional default value.
            Defaults to None.

    Returns:
        Optional[List[bytes]]: A list of the directives.
    """
    value = find(b'clear-site-data', headers)
    return default if value is None else _parse_comma_separated_list(value)


# Connection

_PARSERS[b'connection'] = _Parser(_pass_through, _MergeType.NONE)


def connection(
        headers: Headers,
        *,
        default: Optional[bytes] = None
) -> Optional[bytes]:
    """The Connection general header controls whether or not the network
    connection stays open after the current transaction finishes. If the value
    sent is keep-alive, the connection is persistent and not closed, allowing
    for subsequent requests to the same server to be done.

    Args:
        headers (Headers): The headers.
        default (Optional[bytes], optional): An optional default value. Defaults
            to None.

    Returns:
        Optional[bytes]: The value
    """
    value = find(b'connection', headers)
    return default if value is None else value

# Content-Disposition


def _parse_content_disposition(
        value: bytes
) -> Tuple[bytes, Optional[Mapping[bytes, bytes]]]:
    media_type, sep, rest = value.partition(b';')
    parameters = {
        first.strip(): rest.strip(b'"')
        for first, sep, rest in [x.partition(b'=') for x in rest.split(b';')] if first
    } if sep == b';' else None

    return media_type, parameters


_PARSERS[b'content-disposition'] = _Parser(
    _parse_content_disposition, _MergeType.NONE)


def content_disposition(
        headers: Headers,
        *,
        default: Optional[Tuple[bytes, Optional[Mapping[bytes, bytes]]]] = None
) -> Optional[Tuple[bytes, Optional[Mapping[bytes, bytes]]]]:
    """Returns the content type if any otherwise None

    Args:
        headers (Headers): The headers.
        default (Optional[Tuple[bytes, Optional[Mapping[bytes, bytes]]]], optional): An
            optional default. Defaults to None.

    Returns:
        Optional[Tuple[bytes, Optional[Mapping[bytes, bytes]]]]: A tuple of the
            media type and a mapping of the parameters.
    """
    value = find(b'content-disposition', headers)
    return default if value is None else _parse_content_disposition(value)

# Content-Encoding


def _parse_content_encoding(value: bytes, *, add_identity: bool = False) -> List[bytes]:
    encodings = value.split(b', ')

    if add_identity and b'identity' not in encodings:
        encodings.append(b'identity')

    return encodings


_PARSERS[b'content-encoding'] = _Parser(
    _parse_content_encoding, _MergeType.NONE)


def content_encoding(
        headers: Headers,
        *,
        add_identity: bool = False,
        default: Optional[List[bytes]] = None
) -> Optional[List[bytes]]:
    """Returns the content encodings in a list or None if they were not
    specified.

    Args:
        headers (Headers): The headers.
        add_identity (bool, optional): If True ensures the 'identity' encoding
            is included. Defaults to False.
        default (Optional[List[bytes]], optional): An optional default value.
            Defaults to None.

    Returns:
        Optional[List[bytes]]: The list of content encodings.
    """
    value = find(b'content-encoding', headers)
    return default if value is None else _parse_content_encoding(
        value,
        add_identity=add_identity
    )


# Content-Language

_PARSERS[b'content-language'] = _Parser(
    _parse_comma_separated_list, _MergeType.NONE)


def content_language(
        headers: Headers,
        *,
        default: Optional[List[bytes]] = None
) -> Optional[List[bytes]]:
    """The Content-Language entity header is used to describe the language(s)
    intended for the audience, so that it allows a user to differentiate
    according to the users' own preferred language.

    Args:
        headers (Headers): The headers.
        default (Optional[List[bytes]], optional): An optional default value.
            Defaults to None.

    Returns:
        Optional[List[bytes]]: The language.
    """
    value = find(b'content-language', headers)
    return default if value is None else _parse_comma_separated_list(value)


# Content-Length

_PARSERS[b'content-length'] = _Parser(_parse_int, _MergeType.NONE)


def content_length(
        headers: Headers,
        *,
        default: Optional[int] = None
) -> Optional[int]:
    """[summary]

    Args:
        headers (Headers): The headers.
        default (Optional[int], optional): An optional default value. Defaults
            to None.

    Returns:
        Optional[int]: The length as an integer, or the default.
    """
    value = find(b'content-length', headers)
    return default if value is None else _parse_int(value)

# Content-Location


_PARSERS[b'content-location'] = _Parser(_pass_through, _MergeType.NONE)


def content_location(
        headers: Headers,
        *,
        default: Optional[bytes] = None
) -> Optional[bytes]:
    """The Content-Location header indicates an alternate location for the
    returned data. The principal use is to indicate the URL of a resource
    transmitted as the result of content negotiation.

    Args:
        headers (Headers): The headers.
        default (Optional[bytes], optional): An optional default value. Defaults
            to None.

    Returns:
        Optional[bytes]: The location, or the default.
    """
    return find(b'content-location', headers, default=default)

# Content-Range


def _parse_content_range(
        value: bytes
) -> Tuple[bytes, Optional[Tuple[int, int]], Optional[int]]:
    unit, _, rest = value.strip().partition(b' ')
    range_, _, size_ = rest.strip().partition(b'/')
    if range_ == b'*':
        from_to = None
    else:
        start, _, end = range_.partition(b'-')
        from_to = (int(start), int(end))
    size = None if size_.strip() == b'*' else int(size_)
    return unit, from_to, size


_PARSERS[b'content-range'] = _Parser(_parse_content_range, _MergeType.NONE)


def content_range(
        headers: Headers,
        *,
        default: Optional[Tuple[bytes, Optional[Tuple[int, int]], Optional[int]]] = None,
) -> Optional[Tuple[bytes, Optional[Tuple[int, int]], Optional[int]]]:
    """The Content-Range response HTTP header indicates where in a full body
    message a partial message belongs.

    Args:
        headers (Headers): The headers.
        default (Optional[Tuple[bytes, Optional[Tuple[int, int]], Optional[int]]], optional): An
            optional default value. Defaults to None.

    Returns:
        Optional[Tuple[bytes, Optional[Tuple[int, int]], Optional[int]]]: The
            content-range header if found, or the default.
    """
    value = find(b'content-range', headers)
    return default if value is None else _parse_content_range(value)

# Content-Security-Policy


def _parse_content_security_policy(value: bytes) -> List[Tuple[bytes, List[bytes]]]:
    return [
        (directive.strip(), args.strip().split(b' '))
        for directive, sep, args in [
            policy_directive.strip().partition(b' ')
            for policy_directive in value.strip().split(b';')
        ]
        if sep
    ]


_PARSERS[b'content-security-policy'] = _Parser(
    _parse_content_security_policy, _MergeType.CONCAT)


def content_security_policy(
        headers: Headers,
        *,
        default: Optional[List[Tuple[bytes, List[bytes]]]] = None
) -> Optional[List[Tuple[bytes, List[bytes]]]]:
    """The HTTP Content-Security-Policy response header allows web site
    administrators to control resources the user agent is allowed to load for a
    given page. With a few exceptions, policies mostly involve specifying server
    origins and script endpoints. This helps guard against cross-site scripting
    attacks (XSS).

    Args:
        headers (Headers): The headers
        default (Optional[List[Tuple[bytes, List[bytes]]]], optional): An
            optional default. Defaults to None.

    Returns:
        Optional[List[Tuple[bytes, List[bytes]]]]: The policy or the default.
    """
    value = find(b'content-security-policy', headers)
    return default if value is None else _parse_content_security_policy(value)

# Content-Security-Policy-Report-Only


_PARSERS[b'content-security-policy-report-only'] = _Parser(
    _parse_content_security_policy,
    _MergeType.CONCAT
)


def content_security_policy_report_only(
        headers: Headers,
        *,
        default: Optional[List[Tuple[bytes, List[bytes]]]] = None
) -> Optional[List[Tuple[bytes, List[bytes]]]]:
    """The HTTP Content-Security-Policy-Report-Only response header allows web
    developers to experiment with policies by monitoring (but not enforcing)
    their effects. These violation reports consist of JSON documents sent via an
    HTTP POST request to the specified URI.

    Args:
        headers (Headers): The headers.
        default (Optional[List[Tuple[bytes, List[bytes]]]], optional): An
            optional default value. Defaults to None.

    Returns:
        Optional[List[Tuple[bytes, List[bytes]]]]: The policy, or the default.
    """
    value = find(b'content-security-policy-report-only', headers)
    return default if value is None else _parse_content_security_policy(value)

# Content-Type


def _parse_content_type(
        value: bytes
) -> Tuple[bytes, Optional[Mapping[bytes, bytes]]]:
    media_type, sep, rest = value.partition(b';')
    parameters = {
        first.strip(): rest.strip()
        for first, sep, rest in [x.partition(b'=') for x in rest.split(b';')] if first
    } if sep == b';' else None

    return media_type, parameters


_PARSERS[b'content-type'] = _Parser(_parse_content_type, _MergeType.NONE)


def content_type(
        headers: Headers,
        *,
        default: Optional[Tuple[bytes, Optional[Mapping[bytes, bytes]]]] = None
) -> Optional[Tuple[bytes, Optional[Mapping[bytes, bytes]]]]:
    """Returns the content type if any otherwise None

    Args:
        headers (Headers): The headers
        default (Optional[Tuple[bytes, Optional[Mapping[bytes, bytes]]]], optional): An
            optional default value. Defaults to None.

    Returns:
        Optional[Tuple[bytes, Optional[Mapping[bytes, bytes]]]]: A tuple of the
            media type and a mapping of the parameters or the default if absent.
    """
    value = find(b'content-type', headers)
    return default if value is None else _parse_content_type(value)

# Cookie


def _parse_cookie(value: bytes) -> Mapping[bytes, List[bytes]]:
    cookies: MutableMapping[bytes, List[bytes]] = dict()
    for name, content in decode_cookies(value).items():
        cookies.setdefault(name, []).extend(content)
    return cookies


_PARSERS[b'cookie'] = _Parser(_parse_cookie, _MergeType.EXTEND)


def cookie(headers: Headers) -> Mapping[bytes, List[bytes]]:
    """Returns the cookies as a name-value mapping.

    Args:
        headers (Headers): The headers.

    Returns:
        Mapping[bytes, List[bytes]]: The cookies as a name-value mapping.
    """
    cookies: MutableMapping[bytes, List[bytes]] = dict()
    for value in find_all(b'cookie', headers):
        for name, content in _parse_cookie(value).items():
            cookies.setdefault(name, []).extend(content)
    return cookies

# Cross-Origin-Resource-Policy


_PARSERS[b'cross-origin-resource-policy'] = _Parser(
    _pass_through,
    _MergeType.NONE
)


def cross_origin_resource_policy(
        headers: Headers,
        *,
        default: Optional[bytes] = None
) -> Optional[bytes]:
    """The HTTP Cross-Origin-Resource-Policy response header conveys a desire
    that the browser blocks no-cors cross-origin/cross-site requests to the
    given resource.

    Args:
        headers (Headers): The headers.
        default (Optional[bytes], optional): An optional default value. Defaults
            to None.

    Returns:
        Optional[bytes]: The policy if present or the default.
    """
    value = find(b'cross-origin-resource-policy', headers)
    return default if value is None else value


# Date

_PARSERS[b'date'] = _Parser(_parse_date, _MergeType.NONE)


def date(
        headers: Headers,
        *,
        default: Optional[datetime] = None
) -> Optional[datetime]:
    """The Date general HTTP header contains the date and time at which the
    message was originated.

    Args:
        headers (Headers): The headers.
        default (Optional[datetime], optional): An optional default value.
            Defaults to None.

    Returns:
        Optional[datetime]: The date and time at which the message was originated
    """
    return find_date(b'date', headers, default=default)

# DNT


_PARSERS[b'DNT'] = _Parser(_parse_int, _MergeType.NONE)


def dnt(
        headers: Headers,
        *,
        default: Optional[int] = None
) -> Optional[int]:
    """The DNT (Do Not Track) request header indicates the user's tracking
    preference. It lets users indicate whether they would prefer privacy rather
    than personalized content.

    Args:
        headers (Headers): The headers.
        default (Optional[int], optional): An optional default value. Defaults
            to None.

    Returns:
        Optional[int]: 0 for allow tracking, 1 for deny tracking or the default.
    """
    value = find(b'DNT', headers)
    return default if value is None else _parse_int(value)

# DPR


_PARSERS[b'DPR'] = _Parser(_parse_float, _MergeType.NONE)


def dpr(
        headers: Headers,
        *,
        default: Optional[float] = None
) -> Optional[float]:
    """The DPR header is a Client Hints headers which represents the client
    device pixel ratio (DPR), which is the the number of physical device pixels
    corresponding to every CSS pixel.

    Args:
        headers (Headers): The headers.
        default (Optional[float], optional): An optional default value. Defaults
            to None.

    Returns:
        Optional[float]: The device pixel ratio if present, or the default value.
    """
    value = find(b'DPR', headers)
    return default if value is None else _parse_float(value)

# Device-Memory


_PARSERS[b'device-memory'] = _Parser(_parse_float, _MergeType.NONE)


def device_memory(
        headers: Headers,
        *,
        default: Optional[float] = None
) -> Optional[float]:
    """The Device-Memory header is a Device Memory API header that works like
    Client Hints header which represents the approximate amount of RAM client
    device has.

    Args:
        headers (Headers): The headers
        default (Optional[float], optional): An optional default value.
            Defaults to None.

    Returns:
        Optional[float]: The device memory
    """
    value = find(b'device-memory', headers)
    return default if value is None else _parse_float(value)


# Expect

_PARSERS[b'expect'] = _Parser(_pass_through, _MergeType.NONE)


def expect(
        headers: Headers,
        *,
        default: Optional[bytes] = None
) -> Optional[bytes]:
    """Returns the expect header

    Args:
        headers (Headers): The headers.
        default (Optional[bytes], optional): An optional default value.
            Defaults to None.

    Returns:
        Optional[bytes]: The expect directive if present, or the default value.
    """
    value = find(b'expect', headers)
    return default if value is None else value

# Expires


_PARSERS[b'expires'] = _Parser(_parse_date, _MergeType.NONE)


def expires(
        headers: Headers,
        *,
        default: Optional[datetime] = None
) -> Optional[datetime]:
    """The Expires header contains the date/time after which the response is
    considered stale.

    Args:
        headers (Headers): The headers.
        default (Optional[datetime], optional): An optional default value.
            Defaults to None.

    Returns:
        Optional[datetime]: The date/time after which the response is considered
            stale, or the default value.
    """
    return find_date(b'expires', headers, default=default)

# From


_PARSERS[b'from'] = _Parser(_pass_through, _MergeType.NONE)

# Host


def _parse_host(value: bytes) -> Tuple[bytes, Optional[int]]:
    host_, sep, port = value.partition(b':')
    return (host_, None) if not sep else (host_, int(port))


_PARSERS[b'host'] = _Parser(_parse_host, _MergeType.NONE)


def host(
        headers: Headers,
        *,
        default: Optional[Tuple[bytes, Optional[int]]] = None
) -> Optional[Tuple[bytes, Optional[int]]]:
    """Returns the host header as a name, port tuple

    Args:
        headers (Headers): The headers.
        default (Optional[Tuple[bytes, Optional[int]]], optional): An optional
            default value. Defaults to None.

    Returns:
        Optional[Tuple[bytes, Optional[int]]]: The host as a name, port tuple.
    """
    value = find(b'host', headers)
    return default if value is None else _parse_host(value)

# If-Modified-Since


_PARSERS[b'if-modified-since'] = _Parser(_parse_date, _MergeType.NONE)


def if_modified_since(
        headers: Headers,
        *,
        default: Optional[datetime] = None
) -> Optional[datetime]:
    """The If-Modified-Since request HTTP header makes the request conditional:
    the server will send back the requested resource, with a 200 status, only if
    it has been last modified after the given date. If the request has not been
    modified since, the response will be a 304 without any body; the
    Last-Modified response header of a previous request will contain the date of
    last modification. Unlike If-Unmodified-Since, If-Modified-Since can only be
    used with a GET or HEAD.

    Args:
        headers (Headers): The headers
        default (Optional[datetime], optional): [description]. Defaults to None.

    Returns:
        Optional[datetime]: The timestamp if present, otherwise the default
            value.
    """
    return find_date(b'if-modified-since', headers, default=default)

# Last-Modified


_PARSERS[b'last-modified'] = _Parser(_parse_date, _MergeType.NONE)


def last_modified(
        headers: Headers,
        *,
        default: Optional[datetime] = None
) -> Optional[datetime]:
    """The Last-Modified response HTTP header contains the date and time at
    which the origin server believes the resource was last modified. It is used
    as a validator to determine if a resource received or stored is the same.
    Less accurate than an ETag header, it is a fallback mechanism. Conditional
    requests containing If-Modified-Since or If-Unmodified-Since headers make
    use of this field.

    Args:
        headers (Headers): The headers.
        default (Optional[datetime], optional): An optional default value.
            Defaults to None.

    Returns:
        Optional[datetime]: The timestamp if present, otherwise the default
            value.
    """
    return find_date(b'last-modified', headers, default=default)


# Location

_PARSERS[b'location'] = _Parser(_pass_through, _MergeType.NONE)


def location(
        headers: Headers,
        *,
        default: Optional[bytes] = None
) -> Optional[bytes]:
    """The Location response header indicates the URL to redirect a page to. It
    only provides a meaning when served with a 3xx (redirection) or 201
    (created) status response.

    Args:
        headers (Headers): The headers.
        default (Optional[bytes], optional): An optional default value. Defaults
            to None.

    Returns:
        Optional[bytes]: The redirect location
    """
    value = find(b'location', headers)
    return default if value is None else value

# Origin


_PARSERS[b'origin'] = _Parser(_pass_through, _MergeType.NONE)


def origin(
        headers: Headers,
        *,
        default: Optional[bytes] = None
) -> Optional[bytes]:
    """The Origin request header indicates where a fetch originates from. It
    doesn't include any path information, but only the server name. It is sent
    with CORS requests, as well as with POST requests. It is similar to the
    Referer header, but, unlike this header, it doesn't disclose the whole path.

    Args:
        headers (Headers): The headers.
        default (Optional[bytes], optional): An optional default value.
            Defaults to None.

    Returns:
        Optional[bytes]: The origin if present, otherwise the default value.
    """
    value = find(b'origin', headers)
    return default if value is None else value

# Proxy-Authorisation


def proxy_authorization(
        headers: Headers,
        *,
        default: Optional[Tuple[bytes, bytes]] = None
) -> Optional[Tuple[bytes, bytes]]:
    """The HTTP Proxy-Authorization request header contains the credentials to
    authenticate a user agent to a proxy server, usually after the server has
    responded with a 407 Proxy Authentication Required status and the
    Proxy-Authenticate header.

    Args:
        headers (Headers): The headers.
        default (Optional[Tuple[bytes, bytes]], optional): An optional default
            value. Defaults to None.

    Returns:
        Optional[Tuple[bytes, bytes]]: The type and credentials.
    """
    value = find(b'proxy-authorization', headers)
    return default if value is None else _parse_authorization(value)

# Referer


_PARSERS[b'referer'] = _Parser(_pass_through, _MergeType.NONE)


def referer(
        headers: Headers,
        *,
        default: Optional[bytes] = None
) -> Optional[bytes]:
    """The Referer request header contains the address of the previous web page
    from which a link to the currently requested page was followed. The Referer
    header allows servers to identify where people are visiting them from and
    may use that data for analytics, logging, or optimized caching, for example.

    Args:
        headers (Headers): The headers.
        default (Optional[bytes], optional): An optional default value. Defaults
            to None.

    Returns:
        Optional[bytes]: The referer if present; otherwise the default value.
    """
    value = find(b'referer', headers)
    return default if value is None else value

# Server


_PARSERS[b'server'] = _Parser(_pass_through, _MergeType.NONE)


def server(
        headers: Headers,
        *,
        default: Optional[bytes] = None
) -> Optional[bytes]:
    """The Server header contains information about the software used by the
    origin server to handle the request.

    Args:
        headers (Headers): The headers.
        default (Optional[bytes], optional): An optional default value.
            Defaults to None.

    Returns:
        Optional[bytes]: The product directive
    """
    value = find(b'server', headers)
    return default if value is None else value

# Set-Cookie


_PARSERS[b'set-cookie'] = _Parser(decode_set_cookie, _MergeType.APPEND)


def set_cookie(headers: Headers) -> Mapping[bytes, List[Mapping[str, Any]]]:
    """Returns the cookies as a name-value mapping.

    Args:
        headers (Headers): The headers.

    Returns:
        Mapping[bytes, List[Mapping[str, Any]]]: The cookies as a name-value
            mapping.
    """
    set_cookies: MutableMapping[bytes, List[Mapping[str, Any]]] = dict()
    for header in find_all(b'set-cookie', headers):
        decoded = decode_set_cookie(header)
        set_cookies.setdefault(decoded['name'], []).append(decoded)
    return set_cookies


# Vary

def _parse_vary(value: bytes) -> List[bytes]:
    return value.split(b', ')


_PARSERS[b'vary'] = _Parser(_parse_vary, _MergeType.NONE)


def vary(
        headers: Headers,
        *,
        default: Optional[List[bytes]] = None
) -> Optional[List[bytes]]:
    """Returns the vary header value as a list of headers.

    Args:
        value (bytes): The headers.
        default (Optional[List[bytes]], optional): An optional default value.
            Defaults to None.

    Returns:
        Optional[List[bytes]]: A list of the vary headers if present; otherwise
            the default value.
    """
    value = find(b'vary', headers)
    return default if value is None else _parse_vary(value)


def collect(headers: Headers) -> Mapping[bytes, Any]:
    """Collect all headers into a mapping

    Args:
        headers (Headers): The headers

    Returns:
        Mapping[bytes, Any]: A mapping of the parsed headers
    """
    collection: MutableMapping[bytes, Any] = dict()
    for name, value in headers:
        parser = _PARSERS.get(name, _DEFAULT_PARSER)
        if parser.merge_type == _MergeType.APPEND:
            result = parser.parse(value)
            collection.setdefault(name, []).append(result)
        elif parser.merge_type == _MergeType.CONCAT:
            result = parser.parse(value)
            collection.setdefault(name, []).extend(result)
        elif parser.merge_type == _MergeType.EXTEND:
            result = parser.parse(value)
            dct = collection.setdefault(name, dict())
            for k, v in result.items():
                dct.setdefault(k, []).extend(v)
        else:
            collection[name] = parser.parse(value)
    return collection



"""RFC 850"""

from datetime import datetime, timezone
import re
from typing import Optional

# 'Mon, 23-Mar-20 07:36:36 GMT'
DATE_PATTERN = re.compile(
    r'^(Mon|Tue|Wed|Thu|Fri|Sat|Sun), (\d{2})-(Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)-(\d{2}) (\d{2}):(\d{2}):(\d{2}) GMT$'
)
DAY_NAMES = ("Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun")
MONTH_NAMES = ("Jan", "Feb", "Mar", "Apr", "May", "Jun",
               "Jul", "Aug", "Sep", "Oct", "Nov", "Dec")


def try_parse_date(value: str) -> Optional[datetime]:
    """Parse a date according to RFC 850.

    e.g. 'Mon, 23-Mar-20 07:36:36 GMT'

    Args:
        value (str): The string to parse

    Raises:
        ParseError: If the string cannot be parsed to a date

    Returns:
        datetime: The parsed datetime
    """
    matches = DATE_PATTERN.match(value)
    if not matches:
        return None
    _day_of_week, day, month, year, hour, minute, second = matches.groups()
    return datetime(
        2000 + int(year),
        1 + MONTH_NAMES.index(month),
        int(day),
        int(hour),
        int(minute),
        int(second),
        tzinfo=timezone.utc
    )


def format_date(value: datetime) -> str:
    """Format a date according to RFC 850

    e.g. 'Mon, 23-Mar-20 07:36:36 GMT'

    Args:
        value (datetime): The datetime to format

    Returns:
        str: The formatted datetime
    """
    time_tuple = value.utctimetuple()
    return "{day}, {mday:02d}-{mon}-{year:02d} {hour:02d}:{min:02d}:{sec:02d} GMT".format(
        day=DAY_NAMES[time_tuple.tm_wday],
        mday=time_tuple.tm_mday,
        mon=MONTH_NAMES[time_tuple.tm_mon - 1],
        year=2000-time_tuple.tm_year,
        hour=time_tuple.tm_hour,
        min=time_tuple.tm_min,
        sec=time_tuple.tm_sec
    )



"""Date helpers"""

from datetime import datetime
from typing import Optional

import bareutils.dates.rfc_7231 as rfc_7231
import bareutils.dates.rfc_850 as rfc_850


def try_parse_date(value: str) -> Optional[datetime]:
    """Try to parse a date

    Args:
        value (str): The string to parse

    Returns:
        Optional[datetime]: The parsed date or None if the date could not be
            parsed.
    """
    result = rfc_7231.try_parse_date(value)
    if result:
        return result
    result = rfc_850.try_parse_date(value)
    if result:
        return result
    return None


def parse_date(value: str) -> datetime:
    """Parse a date

    Args:
        value (str): The string to parse

    Raises:
        ValueError: If the date could not be parsed

    Returns:
        datetime: The parsed datetime
    """
    result = try_parse_date(value)
    if not result:
        raise ValueError()
    return result



"""Cookies"""

from datetime import datetime, timedelta
from typing import (
    Any,
    Dict,
    List,
    Mapping,
    MutableMapping,
    Optional,
    Union
)

from .dates import parse_date
from .dates.rfc_7231 import format_date


def encode_set_cookie(
        name: bytes,
        value: bytes,
        *,
        expires: Optional[datetime] = None,
        max_age: Optional[Union[int, timedelta]] = None,
        path: Optional[bytes] = None,
        domain: Optional[bytes] = None,
        secure: bool = False,
        http_only: bool = False,
        same_site: Optional[bytes] = None
) -> bytes:
    """Encode set-cookie

    Args:
        name (bytes): The cookie name
        value (bytes): The cookie value
        expires (Optional[datetime], optional): The time the cookie expires.
            Defaults to None.
        max_age (Optional[Union[int, timedelta]], optional): The maximum age of
            the cookie in seconds. Defaults to None.
        path (Optional[bytes], optional): The cookie path. Defaults to None.
        domain (Optional[bytes], optional): The cookie domain. Defaults to None.
        secure (bool, optional): Indicates if the cookie is restricted to https.
            Defaults to False.
        http_only (bool, optional): Indicates if the cookie is available to the
            API. Defaults to False.
        same_site (Optional[bytes], optional): CORS directive. Defaults to None.

    Raises:
        RuntimeError: Raised if the __Secure- or __Host- was used without secure
        ValueError: If a date could not be parsed.

    Returns:
        bytes: The set-cookie header
    """
    if name.startswith(b'__Secure-') or name.startswith(b'__Host-') and not secure:
        raise RuntimeError(
            'Keys starting __Secure- or __Host- require the secure directive'
        )

    set_cookie = name + b'=' + value

    if expires is not None:
        set_cookie += b'; Expires=' + format_date(expires).encode()

    if max_age is not None:
        if isinstance(max_age, timedelta):
            set_cookie += b'; Max-Age=' + \
                str(int(max_age.total_seconds())).encode()
        else:
            set_cookie += b'; Max-Age=' + str(int(max_age)).encode()

    if domain is not None:
        set_cookie += b'; Domain=' + domain

    if path is not None:
        set_cookie += b'; Path=' + path

    if secure:
        set_cookie += b'; Secure'

    if http_only:
        set_cookie += b'; HttpOnly'

    if same_site is not None:
        set_cookie += b'; SameSite=' + same_site

    return set_cookie


def decode_set_cookie(set_cookie: bytes) -> Mapping[str, Any]:
    """Decode a set-cookie header into a dictionary.

    The `max-age` value is represented as a `datatime.timedelta`.
    The `expires` value is represented as a `datetine.datetime`.
    The `secure` value is represented as a `bool`.

    Args:
        set_cookie (bytes): The set-cookie header

    Returns:
        Mapping[str, Any]: A dictionary of the values
    """
    i = iter(set_cookie.split(b';'))
    key, value = next(i).split(b'=', maxsplit=2)
    result: Dict[str, Any] = {'name': key, 'value': value}
    for item in i:
        key, _, value = item.partition(b'=')
        name = key.lower().strip().decode('ascii')
        if name == 'secure':
            result['secure'] = True
        elif name == 'httponly':
            result['http_only'] = True
        elif name == 'expires':
            result['expires'] = parse_date(value.decode('ascii'))
        elif name == 'max-age':
            result['max_age'] = timedelta(seconds=int(value))
        elif name == 'samesite':
            result['same_site'] = value.decode('ascii')
        else:
            result[name] = value
    return result


def encode_cookies(cookies: Mapping[bytes, List[bytes]]) -> bytes:
    """Encode the cookie header

    Args:
        cookies (Mapping[bytes, List[bytes]]): The cookies

    Returns:
        bytes: The cookie header
    """
    return b'; '.join(
        name + b'=' + value
        for name, values in cookies.items()
        for value in values
    )


def decode_cookies(cookies: bytes) -> Mapping[bytes, List[bytes]]:
    """Decode a cookie header

    Args:
        cookies (bytes): The header

    Returns:
        Mapping[bytes, List[bytes]]: The cookies
    """
    result: MutableMapping[bytes, List[bytes]] = dict()
    for morsel in cookies.rstrip(b'; ').split(b'; '):
        name, _, value = morsel.partition(b'=')
        result.setdefault(name, []).append(value)
    return result


def make_cookie(
        key: bytes,
        value: bytes,
        *,
        expires: Optional[Union[datetime, timedelta]] = None,
        path: Optional[bytes] = None,
        domain: Optional[bytes] = None,
        secure: bool = False,
        http_only: bool = False,
        same_site: Optional[bytes] = None
) -> bytes:
    """Make a set-cookie header

    Args:
        key (bytes): The cookie name
        value (bytes): The cookie value
        expires (Optional[Union[datetime, timedelta]], optional): The expiry
            time of the cookie. Defaults to None.
        path (Optional[bytes], optional): The cookie path. Defaults to None.
        domain (Optional[bytes], optional): The cookie domain. Defaults to None.
        secure (bool, optional): Indicates if the cookie is restricted to https.
            Defaults to False.
        http_only (bool, optional): Indicates if the cookie is available to the
            API. Defaults to False.
        same_site (Optional[bytes], optional): CORS directive. Defaults to None.

    Returns:
        bytes: The set-cookie header
    """
    return encode_set_cookie(
        key,
        value,
        expires=expires if isinstance(expires, datetime) else None,
        max_age=expires if isinstance(expires, (int, timedelta)) else None,
        path=path,
        domain=domain,
        secure=secure,
        http_only=http_only,
        same_site=same_site
    )


def make_expired_cookie(key: bytes, path: bytes = b'/') -> bytes:
    """Make an expired cookie

    Args:
        key (bytes): The cookie name
        path (bytes, optional): The cookie path. Defaults to b'/'.

    Returns:
        bytes: [description]
    """
    return make_cookie(key, b'', expires=timedelta(seconds=0), path=path)



"""Exports from bareUtils"""

from .streaming import (
    text_reader,
    text_writer,
    bytes_reader,
    bytes_writer
)

from .compression import (
    make_gzip_compressobj,
    make_deflate_compressobj,
    make_compress_compressobj,
    compression_writer_adapter,
    compression_writer,
    make_compress_decompressobj,
    make_deflate_decompressobj,
    make_gzip_decompressobj,
    CompressionMiddleware,
    make_default_compression_middleware
)

from .responses import (
    bytes_response,
    text_response,
    json_response
)

from .cookies import (
    encode_set_cookie,
    decode_set_cookie,
    encode_cookies,
    decode_cookies
)

__all__ = [
    "text_writer",
    "text_reader",
    "bytes_writer",
    "bytes_reader",

    'make_gzip_compressobj',
    'make_deflate_compressobj',
    'make_compress_compressobj',
    'compression_writer_adapter',
    'compression_writer',
    'make_compress_decompressobj',
    'make_deflate_decompressobj',
    'make_gzip_decompressobj',
    'CompressionMiddleware',
    'make_default_compression_middleware',

    "bytes_response",
    "text_response",
    "json_response",

    "encode_set_cookie",
    "decode_set_cookie",
    "encode_cookies",
    "decode_cookies"
]



"""Responses"""

import json
from typing import Any, Callable, Optional

from baretypes import Headers, HttpResponse, PushResponses

from .streaming import bytes_writer


def bytes_response(
        status: int,
        headers: Optional[Headers],
        buf: bytes,
        content_type: bytes = b'application/octet-stream',
        chunk_size: int = -1,
        pushes: Optional[PushResponses] = None
) -> HttpResponse:
    """A helper function to create a bytes response.

    Args:
        status (int): The HTTP status code.
        headers (Optional[Headers]): The HTTP headers.
        buf (bytes): The data to send.
        content_type (bytes, optional): The content type of the data. Defaults
            to b'application/octet-stream'.
        chunk_size (int, optional): The size of each chunk to send or -1 to send
            as a single chunk. Defaults to -1.
        pushes (Optional[PushResponses], optional): Optional posh responses for
            HTTP/2. Defaults to None.

    Returns:
        HttpResponse: The HTTP response.
    """
    headers = [] if headers is None else list(headers)

    headers.append((b'content-type', content_type))
    if chunk_size == -1:
        headers.append((b'content-length', str(len(buf)).encode('ascii')))

    return status, headers, bytes_writer(buf, chunk_size), pushes


def text_response(
        status: int,
        headers: Optional[Headers],
        text: str,
        encoding: str = 'utf-8',
        content_type: bytes = b'text/plain',
        chunk_size: int = -1,
        pushes: Optional[PushResponses] = None
) -> HttpResponse:
    """A helper function to create a text response.

    Args:
        status (int): The HTTP status code.
        headers (Optional[Headers]): The HTTP headers.
        text (str): The text to send.
        encoding (str, optional): [description]. Defaults to 'utf-8'.
        content_type (bytes, optional): The content type. Defaults to b'text/plain'.
        chunk_size (int, optional): The size of each chunk to send or -1 to send
            as a single chunk. Defaults to -1.
        pushes (Optional[PushResponses], optional): Optional posh responses for
            HTTP/2. Defaults to None.

    Returns:
        HttpResponse: The HTTP response.
    """
    return bytes_response(status, headers, text.encode(encoding), content_type, chunk_size, pushes)


def json_response(
        status: int,
        headers: Optional[Headers],
        obj: Any,
        content_type: bytes = b'application/json',
        dumps: Callable[[Any], str] = json.dumps,
        chunk_size: int = -1,
        pushes: Optional[PushResponses] = None
) -> HttpResponse:
    """A helper function to send a json response.

    Args:
        status (int): The HTTP status code.
        headers (Optional[Headers]): The HTTP headers.
        obj (Any): The object to send as JSON.
        content_type (bytes, optional): The content type. Defaults to
            b'application/json'.
        dumps (Callable[[Any], str], optional): The function to use to turn the
            object into JSON. Defaults to json.dumps.
        chunk_size (int, optional): The size of each chunk to send or -1 to send
            as a single chunk. Defaults to -1.
        pushes (Optional[PushResponses], optional): Optional posh responses for
            HTTP/2. Defaults to None.

    Returns:
        HttpResponse: The HTTP response.
    """
    return text_response(status, headers, dumps(obj), 'utf-8', content_type, chunk_size, pushes)



"""Functions and constants for dealing with HTTP response status codes.

Attributes:
    CONTINUE (int): 100 - This interim response indicates that everything so far
        is OK and that the client should continue with the request or ignore it if it
        is already finished.
    SWITCHING_PROTOCOL (int): 101 - This code is sent in response to an Upgrade
        request header by the client, and indicates the protocol the server is
        switching to.
    PROCESSING (int): 102 - This code indicates that the server has received and
        is processing the request, but no response is available yet.
    EARLY_HINTS (int): 103 - This status code is primarily intended to be used
        with the Link header to allow the user agent to start preloading
        resources while the server is still preparing a response.
    OK (int): 200 - The request has succeeded. The meaning of a success varies
        depending on the HTTP method: For GET the resource has been fetched and
        is transmitted in the message body. For HEAD the entity headers are in
        the message body. For PUT or POST the resource describing the result of
        the action is transmitted in the message body. For TRACE the message
        body contains the request message as received by the server
    CREATED (int): 201 - The request has succeeded and a new resource has been
        created as a result of it. This is typically the response sent  after a
        POST request, or after some PUT requests.
    ACCEPTED (int): 202 - The request has been received but not yet acted upon.
        It is non-committal, meaning that there is no way in HTTP to later send
        an asynchronous response indicating the outcome of processing the
        request. It is intended for cases where another process or server
        handles the request, or for batch processing.
    NON_AUTHORITATIVE_INFORMATION (int): 203 - This response code means returned
        meta-information set is not exact set as available from the origin
        server, but collected from a local or a third party copy. Except this
        condition, 200 OK response should be preferred instead of this response.
    NO_CONTENT (int): 204 - There is no content to send for this request, but
        the headers may be useful. The user-agent may update its cached headers
        for this resource with the new ones.
    RESET_CONTENT (int): 205 - This response code is sent after accomplishing
        request to tell user agent reset document view which sent this request.
    PARTIAL_CONTENT (int): 206 - This response code is used because of range
        header sent by the client to separate download into multiple streams.
    MULTI_STATUS (int): 207 - A Multi-Status response conveys information about
        multiple resources in situations where multiple status codes might be
        appropriate.
    MULTI_STATUS_DAV (int): 208 - Used inside a DAV: propstat response element
        to avoid enumerating the internal members of multiple bindings to the 
        same collection repeatedly.
    IM_USED (int): 226 - The server has fulfilled a GET request for the
        resource, and the response is a representation of the result of one or 
        more instance-manipulations applied to the current instance.
    MULTIPLE_CHOICE (int): 300 - The request has more than one possible
        response. The user-agent or user should choose one of them. There is no
        standardized way of choosing one of the responses.
    MOVED_PERMANENTLY (int): 301 - This response code means that the URI of the
        requested resource has been changed permanently. Probably, the new URI 
        would be given in the response.
    FOUND (int): 302 - This response code means that the URI of requested
        resource has been changed temporarily. New changes in the URI might be
        made in the future. Therefore, this same URI should be used by the
        client in future requests.
    SEE_OTHER (int): 303 - The server sent this response to direct the client to
        get the requested resource at another URI with a GET request.
    NOT_MODIFIED (int): 304 - This is used for caching purposes. It tells the
        client that the response has not been modified, so the client can
        continue to use the same cached version of the response.
    USE_PROXY (int): 305 - Was defined in a previous version of the HTTP
        specification to indicate that a requested response must be accessed by
        a proxy. It has been deprecated due to security concerns regarding
        in-band configuration of a proxy.
    UNUSED (int): 306 - This response code is no longer used, it is just
        reserved currently. It was used in a previous version of the HTTP 1.1
        specification.
    TEMPORARY_REDIRECT (int): 307 - The server sends this response to direct the
        client to get the requested resource at another URI with same method
        that was used in the prior request. This has the same semantics as the
        302 Found HTTP response code, with the exception that the user agent
        must not change the HTTP method used: If a POST was used in the first
        request, a POST must be used in the second request.
    PERMANENT_REDIRECT (int): 308 - This means that the resource is now
        permanently located at another URI, specified by the Location: HTTP
        Response header. This has the same semantics as the 301 Moved
        Permanently HTTP response code, with the exception that the user agent
        must not change the HTTP method used: If a POST was used in the first
        request, a POST must be used in the second request.
    BAD_REQUEST (int): 400 - This response means that server could not
        understand the request due to invalid syntax.
    UNAUTHORIZED (int): 401 - Although the HTTP standard specifies
        "unauthorized", semantically this response means "unauthenticated". That
        is, the client must authenticate itself to get the requested response.
    PAYMENT_REQUIRED (int): 402 - This response code is reserved for future use.
        Initial aim for creating this code was using it for digital payment
        systems however this is not used currently.
    FORBIDDEN (int): 403 - The client does not have access rights to the
        content, i.e. they are unauthorized, so server is rejecting to give
        proper response. Unlike 401, the client's identity is known to the
        server.
    NOT_FOUND (int): 404 - The server can not find requested resource. In the
        browser, this means the URL is not recognized. In an API, this can also
        mean that the endpoint is valid but the resource itself does not exist.
        Servers may also send this response instead of 403 to hide the existence
        of a resource from an unauthorized client. This response code is
        probably the most famous one due to its frequent occurrence on the web.
    METHOD_NOT_ALLOWED (int): 405 - The request method is known by the server
        but has been disabled and cannot be used. For example, an API may forbid 
        DELETE-ing a resource. The two mandatory methods, GET and HEAD, must
        never be disabled and should not return this error code.
    NOT_ACCEPTABLE (int): 406 - This response is sent when the web server, after
        performing server-driven content negotiation, doesn't find any content
        following the criteria given by the user agent.
    PROXY_AUTHENTICATION_REQUIRED (int): 407 - This is similar to 401 but
        authentication is needed to be done by a proxy.
    REQUEST_TIMEOUT (int): 408 - This response is sent on an idle connection by
        some servers, even without any previous request by the client. It  means
        that the server would like to shut down this unused connection. This
        response is used much more since some browsers, like Chrome, Firefox
        27+, or IE9, use HTTP pre-connection mechanisms to speed up surfing.
        Also note that some servers merely shut down the connection without
        sending this message.
    CONFLICT (int): 409 - This response is sent when a request conflicts with
        the current state of the server.
    GONE (int): 410 - This response would be sent when the requested content has
        been permanently deleted from server, with no forwarding address.
        Clients are expected to remove their caches and links to the resource.
        The HTTP specification intends this status code to be used for
        "limited-time, promotional services". APIs should not feel compelled to
        indicate resources that have been deleted with this status code.
    LENGTH_REQUIRED (int): 411 - Server rejected the request because the
        content-length header field is not defined and the server requires it.
    PRECONDITION_FAILED (int): 412 - The client has indicated preconditions in
        its headers which the server does not meet.
    PAYLOAD_TOO_LARGE (int): 413 - Request entity is larger than limits defined
        by server; the server might close the connection or return an
        retry-after header field.
    URI_TOO_LONG (int): 414 - The URI requested by the client is longer than the
        server is willing to interpret.
    UNSUPPORTED_MEDIA_TYPE (int): 415 - The media format of the requested data
        is not supported by the server, so the server is rejecting the request.
    REQUESTED_RANGE_NOT_SATISFIABLE (int): 416 - The range specified by the
        Range header field in the request can't be fulfilled; it's possible that
        the range is outside the size of the target URI's data.
    EXPECTATION_FAILED (int): 417 - This response code means the expectation
        indicated by the Expect request header field can't be met by the server.
    IM_A_TEAPOT (int): 418 - The server refuses the attempt to brew coffee with
        a teapot.
    MISDIRECTED_REQUEST (int): 421 - The request was directed at a server that
        is not able to produce a response. This can be sent by a server that is 
        not configured to produce responses for the combination of scheme and
        authority that are included in the request URI.
    UNPROCESSABLE_ENTITY (int): 422 - The request was well-formed but was unable
        to be followed due to semantic errors.
    LOCKED (int): 423 - The resource that is being accessed is locked.
    FAILED_DEPENDENCY (int): 424 - The request failed due to failure of a
        previous request.
    TOO_EARLY (int): 425 - Indicates that the server is unwilling to risk
        processing a request that might be replayed.
    UPGRADE_REQUIRED (int): 426 - The server refuses to perform the request
        using the current protocol but might be willing to do so after the
        client upgrades to a different protocol. The server sends an Upgrade
        header in a 426 response to indicate the required protocol(s).
    PRECONDITION_REQUIRED (int): 428 - The origin server requires the request to
        be conditional. Intended to prevent the 'lost update' problem, where a
        client GETs a resource's state, modifies it, and PUTs it back to the
        server, when meanwhile a third party has modified the state on the
        server, leading to a conflict.
    TOO_MANY_REQUESTS (int): 429 - The user has sent too many requests in a
        given amount of time ("rate limiting").
    REQUEST_HEADER_FIELDS_TOO_LARGE (int): 431 - The server is unwilling to
        process the request because its header fields are too large. The request
        MAY be resubmitted after reducing the size of the request header fields.
    UNAVAILABLE_FOR_LEGAL_REASONS (int): 451 - The user requests an illegal
        resource, such as a web page censored by a government.
    INTERNAL_SERVER_ERROR (int): 500 - The server has encountered a situation it
        doesn't know how to handle.
    NOT_IMPLEMENTED (int): 501 - The request method is not supported by the
        server and cannot be handled. The only methods that servers are required 
        to support (and therefore that must not return this code) are GET and
        HEAD.
    BAD_GATEWAY (int): 502 - This error response means that the server, while
        working as a gateway to get a response needed to handle the request, got
        an invalid response.
    SERVICE_UNAVAILABLE (int): 503 - The server is not ready to handle the
        request. Common causes are a server that is down for maintenance or that
        is overloaded. Note that together with this response, a user-friendly
        page explaining the problem should be sent. This responses should be
        used for temporary conditions and the Retry-After: HTTP header should,
        if possible, contain the estimated time before the recovery of the
        service. The webmaster must also take care about the caching-related
        headers that are sent along with this response, as these temporary
        condition responses should usually not be cached.
    GATEWAY_TIMEOUT (int): 504 - This error response is given when the server is
        acting as a gateway and cannot get a response in time.
    HTTP_VERSION_NOT_SUPPORTED (int): 505 - The HTTP version used in the request
        is not supported by the server.
    VARIANT_ALSO_NEGOTIATES (int): 506 - The server has an internal
        configuration error: transparent content negotiation for the request
        results in a circular reference.
    INSUFFICIENT_STORAGE (int): 507 - The server has an internal configuration
        error: the chosen variant resource is configured to engage in
        transparent content negotiation itself, and is therefore not a proper
        end point in the negotiation process.
    LOOP_DETECTED (int): 508 - The server detected an infinite loop while
        processing the request.
    NOT_EXTENDED (int): 510 - Further extensions to the request are required for
        the server to fulfill it.
    NETWORK_AUTHENTICATION_IS_REQUIRED 9int): 511 - The 511 status code
        indicates that the client needs to authenticate to gain network access.
"""


def is_information(code: int) -> bool:
    """Return true if the code is an information HTTP response code.

    Args:
        code (int): The HTTP response code.

    Returns:
        bool: True if the code was informational else false.
    """
    return code >= 100 and code < 200


def is_successful(code: int) -> bool:
    """Return true if the code is a successful HTTP response code.

    Args:
        code (int): The HTTP response code.

    Returns:
        bool: True if the code was successful else false.
    """
    return code >= 200 and code < 300


def is_redirection(code: int) -> bool:
    """Return true if the code is aa redirection HTTP response code.

    Args:
        code (int): The HTTP response code.

    Returns:
        bool: True if the code was a redirection else false.
    """
    return code >= 300 and code < 400


def is_client_error(code: int) -> bool:
    """Return true if the code is a client error HTTP response code.

    Args:
        code (int): The HTTP response code.

    Returns:
        bool: True if the code was a client error else false.
    """
    return code >= 400 and code < 500


def is_server_error(code: int) -> bool:
    """Return true if the code is a server error HTTP response code.

    Args:
        code (int): The HTTP response code.

    Returns:
        bool: True if the code was a server error else false.
    """
    return code >= 500 and code < 600


# Information responses

"""
This interim response indicates that everything so far is OK and that the client should continue with the request or
ignore it if it is already finished.
"""
CONTINUE = 100

"""
This code is sent in response to an Upgrade request header by the client, and indicates the protocol the server is 
switching to.
"""
SWITCHING_PROTOCOL = 101

"""
This code indicates that the server has received and is processing the request, but no response is available yet.
"""
PROCESSING = 102

"""
This status code is primarily intended to be used with the Link header to allow the user agent to start preloading 
resources while the server is still preparing a response.
"""
EARLY_HINTS = 103

# Successful responses

"""
The request has succeeded. The meaning of a success varies depending on the HTTP method:
GET: The resource has been fetched and is transmitted in the message body.
HEAD: The entity headers are in the message body.
PUT or POST: The resource describing the result of the action is transmitted in the message body.
TRACE: The message body contains the request message as received by the server
"""
OK = 200

"""
The request has succeeded and a new resource has been created as a result of it. This is typically the response sent 
after a POST request, or after some PUT requests.
"""
CREATED = 201

"""
The request has been received but not yet acted upon. It is non-committal, meaning that there is no way in HTTP to 
later send an asynchronous response indicating the outcome of processing the request. It is intended for cases where 
another process or server handles the request, or for batch processing.
"""
ACCEPTED = 202

"""
This response code means returned meta-information set is not exact set as available from the origin server, but 
collected from a local or a third party copy. Except this condition, 200 OK response should be preferred instead of 
this response.
"""
NON_AUTHORITATIVE_INFORMATION = 203

"""
There is no content to send for this request, but the headers may be useful. The user-agent may update its cached 
headers for this resource with the new ones.
"""
NO_CONTENT = 204

"""
This response code is sent after accomplishing request to tell user agent reset document view which sent this request.
"""
RESET_CONTENT = 205

"""
This response code is used because of range header sent by the client to separate download into multiple streams.
"""
PARTIAL_CONTENT = 206

"""
A Multi-Status response conveys information about multiple resources in situations where multiple status codes might 
be appropriate.
"""
MULTI_STATUS = 207

"""
Used inside a DAV: propstat response element to avoid enumerating the internal members of multiple bindings to the 
same collection repeatedly.
"""
MULTI_STATUS_DAV = 208

"""
The server has fulfilled a GET request for the resource, and the response is a representation of the result of one or 
more instance-manipulations applied to the current instance.
"""
IM_USED = 226

# Redirection messages

"""
The request has more than one possible response. The user-agent or user should choose one of them. There is no 
standardized way of choosing one of the responses.
"""
MULTIPLE_CHOICE = 300

"""
This response code means that the URI of the requested resource has been changed permanently. Probably, the new URI 
would be given in the response.
"""
MOVED_PERMANENTLY = 301

"""
This response code means that the URI of requested resource has been changed temporarily. New changes in the URI 
might be made in the future. Therefore, this same URI should be used by the client in future requests.
"""
FOUND = 302

"""
The server sent this response to direct the client to get the requested resource at another URI with a GET request.
"""
SEE_OTHER = 303

"""
This is used for caching purposes. It tells the client that the response has not been modified, so the client 
can continue to use the same cached version of the response.
"""
NOT_MODIFIED = 304

"""
Was defined in a previous version of the HTTP specification to indicate that a requested response must be accessed 
by a proxy. It has been deprecated due to security concerns regarding in-band configuration of a proxy.
"""
USE_PROXY = 305

"""
This response code is no longer used, it is just reserved currently. It was used in a previous version of the 
HTTP 1.1 specification.
"""
UNUSED = 306

"""
The server sends this response to direct the client to get the requested resource at another URI with same method 
that was used in the prior request. This has the same semantics as the 302 Found HTTP response code, with the 
exception that the user agent must not change the HTTP method used: If a POST was used in the first request, a 
POST must be used in the second request.
"""
TEMPORARY_REDIRECT = 307

"""
This means that the resource is now permanently located at another URI, specified by the Location: HTTP Response 
header. This has the same semantics as the 301 Moved Permanently HTTP response code, with the exception that the 
user agent must not change the HTTP method used: If a POST was used in the first request, a POST must be used in 
the second request.
"""
PERMANENT_REDIRECT = 308

# Client error responses

"""
This response means that server could not understand the request due to invalid syntax.
"""
BAD_REQUEST = 400

"""
Although the HTTP standard specifies "unauthorized", semantically this response means "unauthenticated". That is, 
the client must authenticate itself to get the requested response.
"""
UNAUTHORIZED = 401

"""
This response code is reserved for future use. Initial aim for creating this code was using it for digital payment 
systems however this is not used currently.
"""
PAYMENT_REQUIRED = 402

"""
The client does not have access rights to the content, i.e. they are unauthorized, so server is rejecting to give 
proper response. Unlike 401, the client's identity is known to the server.
"""
FORBIDDEN = 403

"""
The server can not find requested resource. In the browser, this means the URL is not recognized. In an API, this 
can also mean that the endpoint is valid but the resource itself does not exist. Servers may also send this 
response instead of 403 to hide the existence of a resource from an unauthorized client. This response code is 
probably the most famous one due to its frequent occurence on the web.
"""
NOT_FOUND = 404

"""
The request method is known by the server but has been disabled and cannot be used. For example, an API may forbid 
DELETE-ing a resource. The two mandatory methods, GET and HEAD, must never be disabled and should not return this 
error code.
"""
METHOD_NOT_ALLOWED = 405

"""
This response is sent when the web server, after performing server-driven content negotiation, doesn't find any 
content following the criteria given by the user agent.
"""
NOT_ACCEPTABLE = 406

"""
This is similar to 401 but authentication is needed to be done by a proxy.
"""
PROXY_AUTHENTICATION_REQUIRED = 407

"""
This response is sent on an idle connection by some servers, even without any previous request by the client. It 
means that the server would like to shut down this unused connection. This response is used much more since some 
browsers, like Chrome, Firefox 27+, or IE9, use HTTP pre-connection mechanisms to speed up surfing. Also note that 
some servers merely shut down the connection without sending this message.
"""
REQUEST_TIMEOUT = 408

"""
This response is sent when a request conflicts with the current state of the server.
"""
CONFLICT = 409

"""
This response would be sent when the requested content has been permanently deleted from server, with no forwarding 
address. Clients are expected to remove their caches and links to the resource. The HTTP specification intends this 
status code to be used for "limited-time, promotional services". APIs should not feel compelled to indicate resources 
that have been deleted with this status code.
"""
GONE = 410

"""
Server rejected the request because the Content-Length header field is not defined and the server requires it.
"""
LENGTH_REQUIRED = 411

"""
The client has indicated preconditions in its headers which the server does not meet.
"""
PRECONDITION_FAILED = 412

"""
Request entity is larger than limits defined by server; the server might close the connection or return an Retry-After header field.
"""
PAYLOAD_TOO_LARGE = 413

"""
The URI requested by the client is longer than the server is willing to interpret.
"""
URI_TOO_LONG = 414

"""
The media format of the requested data is not supported by the server, so the server is rejecting the request.
"""
UNSUPPORTED_MEDIA_TYPE = 415

"""
The range specified by the Range header field in the request can't be fulfilled; it's possible that the range is 
outside the size of the target URI's data.
"""
REQUESTED_RANGE_NOT_SATISFIABLE = 416

"""
This response code means the expectation indicated by the Expect request header field can't be met by the server.
"""
EXPECTATION_FAILED = 417

"""
The server refuses the attempt to brew coffee with a teapot.
"""
IM_A_TEAPOT = 418

"""
The request was directed at a server that is not able to produce a response. This can be sent by a server that is 
not configured to produce responses for the combination of scheme and authority that are included in the request URI.
"""
MISDIRECTED_REQUEST = 421

"""
The request was well-formed but was unable to be followed due to semantic errors.
"""
UNPROCESSABLE_ENTITY = 422

"""
The resource that is being accessed is locked.
"""
LOCKED = 423

"""
The request failed due to failure of a previous request.
"""
FAILED_DEPENDENCY = 424

"""
Indicates that the server is unwilling to risk processing a request that might be replayed.
"""
TOO_EARLY = 425

"""
The server refuses to perform the request using the current protocol but might be willing to do so after the client 
upgrades to a different protocol. The server sends an Upgrade header in a 426 response to indicate the required 
protocol(s).
"""
UPGRADE_REQUIRED = 426

"""
The origin server requires the request to be conditional. Intended to prevent the 'lost update' problem, where a 
client GETs a resource's state, modifies it, and PUTs it back to the server, when meanwhile a third party has 
modified the state on the server, leading to a conflict.
"""
PRECONDITION_REQUIRED = 428

"""
The user has sent too many requests in a given amount of time ("rate limiting").
"""
TOO_MANY_REQUESTS = 429

"""
The server is unwilling to process the request because its header fields are too large. The request MAY be 
resubmitted after reducing the size of the request header fields.
"""
REQUEST_HEADER_FIELDS_TOO_LARGE = 431

"""
The user requests an illegal resource, such as a web page censored by a government.
"""
UNAVAILABLE_FOR_LEGAL_REASONS = 451

# Server error responses

"""
The server has encountered a situation it doesn't know how to handle.
"""
INTERNAL_SERVER_ERROR = 500

"""
The request method is not supported by the server and cannot be handled. The only methods that servers are required 
to support (and therefore that must not return this code) are GET and HEAD.
"""
NOT_IMPLEMENTED = 501

"""
This error response means that the server, while working as a gateway to get a response needed to handle the 
request, got an invalid response.
"""
BAD_GATEWAY = 502

"""
The server is not ready to handle the request. Common causes are a server that is down for maintenance or that 
is overloaded. Note that together with this response, a user-friendly page explaining the problem should be sent. 
This responses should be used for temporary conditions and the Retry-After: HTTP header should, if possible, 
contain the estimated time before the recovery of the service. The webmaster must also take care about the 
caching-related headers that are sent along with this response, as these temporary condition responses should 
usually not be cached.
"""
SERVICE_UNAVAILABLE = 503

"""
This error response is given when the server is acting as a gateway and cannot get a response in time.
"""
GATEWAY_TIMEOUT = 504

"""
The HTTP version used in the request is not supported by the server.
"""
HTTP_VERSION_NOT_SUPPORTED = 505

"""
The server has an internal configuration error: transparent content negotiation for the request results in a 
circular reference.
"""
VARIANT_ALSO_NEGOTIATES = 506

"""
The server has an internal configuration error: the chosen variant resource is configured to engage in transparent 
content negotiation itself, and is therefore not a proper end point in the negotiation process.
"""
INSUFFICIENT_STORAGE = 507

"""
The server detected an infinite loop while processing the request.
"""
LOOP_DETECTED = 508

"""
Further extensions to the request are required for the server to fulfill it.
"""
NOT_EXTENDED = 510

"""
The 511 status code indicates that the client needs to authenticate to gain network access.
"""
NETWORK_AUTHENTICATION_IS_REQUIRED = 511



"""Cookies"""

from datetime import datetime, timedelta, timezone
import re
from typing import (
    Any,
    Dict,
    List,
    Mapping,
    MutableMapping,
    Optional,
    Union
)

from baretypes import ParseError

# pylint: disable=line-too-long
# Date: <day-name>, <day> <month> <year> <hour>:<minute>:<second> GMT
DATE_PATTERN = re.compile(
    r'^(Mon|Tue|Wed|Thu|Fri|Sat|Sun), (\d{2}) (Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec) (\d{4}) (\d{2}):(\d{2}):(\d{2}) GMT$'
)
DAY_NAMES = ("Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun")
MONTH_NAMES = ("Jan", "Feb", "Mar", "Apr", "May", "Jun",
               "Jul", "Aug", "Sep", "Oct", "Nov", "Dec")


def try_parse_date(value: str) -> Optional[datetime]:
    """Parse a date according to RFC 7231, section 7.1.1.2: Date

    Args:
        value (str): The string to parse

    Raises:
        ParseError: If the string cannot be parsed to a date

    Returns:
        datetime: The parsed datetime
    """
    matches = DATE_PATTERN.match(value)
    if not matches:
        return None

    _day_of_week, day, month, year, hour, minute, second = matches.groups()
    return datetime(
        int(year),
        1 + MONTH_NAMES.index(month),
        int(day),
        int(hour),
        int(minute),
        int(second),
        tzinfo=timezone.utc
    )


def parse_date(value: str) -> datetime:
    """Parses a date according to RFC 7231, section 7.1.1.2: Date

    Args:
        value (str): The value to parse.

    Raises:
        ValueError: If the format could not be parsed.

    Returns:
        datetime: The parsed date.
    """
    return datetime.strptime(value, '%a, %d %b %Y %H:%M:%S %Z')


def format_date(value: datetime) -> str:
    """Format a date according to RFC 7231, section 7.1.1.2: Date

    Args:
        value (datetime): The datetime to format

    Returns:
        str: The formatted datetime
    """
    time_tuple = value.utctimetuple()
    return "{day}, {mday:02d} {mon} {year:04d} {hour:02d}:{min:02d}:{sec:02d} GMT".format(
        day=DAY_NAMES[time_tuple.tm_wday],
        mday=time_tuple.tm_mday,
        mon=MONTH_NAMES[time_tuple.tm_mon - 1],
        year=time_tuple.tm_year,
        hour=time_tuple.tm_hour,
        min=time_tuple.tm_min,
        sec=time_tuple.tm_sec
    )



"""Exports for compression"""

from .streaming import (
    Compressor,
    make_gzip_compressobj,
    make_deflate_compressobj,
    make_compress_compressobj,
    compression_writer_adapter,
    compression_writer,
    Decompressor,
    make_gzip_decompressobj,
    make_deflate_decompressobj,
    make_compress_decompressobj,
    compression_reader_adapter,
    compression_reader
)

from .middleware import (
    CompressionMiddleware,
    make_default_compression_middleware
)

__all__ = [
    'Compressor',
    'make_gzip_compressobj',
    'make_deflate_compressobj',
    'make_compress_compressobj',
    'compression_writer_adapter',
    'compression_writer',
    'make_gzip_decompressobj',
    'make_deflate_decompressobj',
    'make_compress_decompressobj',
    'compression_reader_adapter',
    'compression_reader',

    'CompressionMiddleware',
    'make_default_compression_middleware'
]



"""
Compression streaming.
"""

from __future__ import annotations
from abc import ABCMeta, abstractmethod
from typing import Optional
import zlib

from baretypes import Content

from ..streaming import bytes_writer, bytes_reader


class Compressor(metaclass=ABCMeta):
    """A class to represent the methods available on a compressor"""

    @abstractmethod
    def compress(self, buf: bytes) -> bytes:
        """Compress a buffer

        Args:
            buf (bytes): The buffer to compress.

        Returns:
            bytes: The compressed buffer.
        """

    def flush(self) -> bytes:
        """Flush the compressor

        Returns:
            bytes: The remaining bytes.
        """


def make_gzip_compressobj() -> Compressor:
    """Make a compressor for 'gzip'

    Returns:
        Compressor: A gzip compressor.
    """
    return zlib.compressobj(9, zlib.DEFLATED, zlib.MAX_WBITS | 16)


def make_deflate_compressobj() -> Compressor:
    """Make a compressor for 'deflate'

    Returns:
        Compressor: A deflate compressor.
    """
    return zlib.compressobj(9, zlib.DEFLATED, -zlib.MAX_WBITS)


def make_compress_compressobj() -> Compressor:
    """Make a compressor for 'compress'

    Note: This is not used by browsers anymore and should be avoided.

    Returns:
        Compressor: A compress compressor.
    """
    return zlib.compressobj(9, zlib.DEFLATED, zlib.MAX_WBITS)


async def compression_writer_adapter(
        writer: Content,
        compressobj: Compressor
) -> Content:
    """Adapts a bytes generator to generated compressed output.

    Args:
        writer (Content): The writer to be adapted.
        compressobj (Compressor): A compressor

    Yields:
        Content: The compressed content as bytes
    """
    async for buf in writer:
        yield compressobj.compress(buf)
    yield compressobj.flush()


def compression_writer(
        buf: bytes,
        compressobj: Compressor,
        chunk_size: int = -1
) -> Content:
    """Create an async iterator for compressed content.

    Args:
        buf (bytes): The bytes to compress.
        compressobj (Compressor): The compressor.
        chunk_size (int, optional): An optional chunk size where -1 indicates no
            chunking. Defaults to -1.

    Returns:
        Content: An async iterator of compressed bytes.
    """
    return compression_writer_adapter(bytes_writer(buf, chunk_size), compressobj)


class Decompressor(metaclass=ABCMeta):
    """A class to represent the methods available on a compressor"""

    @property
    @abstractmethod
    def unused_data(self) -> bytes:
        ...

    @property
    @abstractmethod
    def unconsumed_tail(self) -> bytes:
        ...

    @property
    @abstractmethod
    def eof(self) -> bool:
        ...

    @abstractmethod
    def decompress(self, buf: bytes, max_length: int = 0) -> bytes:
        ...

    @abstractmethod
    def flush(self, length: Optional[int] = None) -> bytes:
        ...

    @abstractmethod
    def copy(self) -> Decompressor:
        ...


def make_gzip_decompressobj() -> Decompressor:
    """Make a compressor for 'gzip'

    Returns:
        Decompressor: A gzip compressor.
    """
    return zlib.decompressobj(zlib.MAX_WBITS | 16)


def make_deflate_decompressobj() -> Decompressor:
    """Make a compressor for 'deflate'

    Returns:
        Decompressor: A deflate compressor.
    """
    return zlib.decompressobj(-zlib.MAX_WBITS)


def make_compress_decompressobj() -> Decompressor:
    """Make a compressor for 'compress'

    Note: This is not used by browsers anymore and should be avoided.

    Returns:
        Decompressor: A compress compressor.
    """
    return zlib.decompressobj(9, zlib.DEFLATED, zlib.MAX_WBITS)


async def compression_reader_adapter(
        reader: Content,
        decompressobj: Decompressor
) -> Content:
    async for item in reader:
        yield decompressobj.decompress(item)
    yield decompressobj.flush()


async def compression_reader(source: Content, decompressobj: Decompressor) -> bytes:
    return await bytes_reader(compression_reader_adapter(source, decompressobj))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['bareutils', 'bareutils.compression', 'bareutils.dates']

package_data = \
{'': ['*']}

install_requires = \
['baretypes>=3,<4']

setup_kwargs = {
    'name': 'bareutils',
    'version': '3.5.0',
    'description': 'Utilities for bareASGI and bareClient',
    'long_description': '# bareutils\n\nUtilities for [bareASGI](https://github.com/rob-blackbourn/bareASGI)\nand [bareClient](https://github.com/rob-blackbourn/bareClient)\n(read the [docs](https://rob-blackbourn.github.io/bareUtils/)).\n\n## Installation\n\nThe package can be installed with pip.\n\n```bash\npip install bareutils\n```\n\nThis is a Python3.7 and later package.\n\nIt has dependencies on:\n\n* [bareTypes](https://github.com/rob-blackbourn/bareTypes)\n',
    'author': 'Rob Blackbourn',
    'author_email': 'rob.blackbourn@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/rob-blackbourn/bareutils',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
